/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Random;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;

import sequoia.topology.ITopology;
import sequoia.tree.IPredictionTree;

public class Master {
	
	public static Random rand;
	public static boolean CDF_LIMIT = true;
	public static int TREENESSLIMIT = 10;
	private static final String CDF_PATH = "cdfs/";
	public static final String CLIQUE_FILE = "data/clique.txt";
	public static int CONST = 10000000; // transform constant
	private static CommandLine cli;
	private static Options options;
	public static final int CDF_ERROR_MAX = 10;
	private static final int PERCENTILE = 60;
	
	// variants of output matrices
	public static enum Output {
		MEASURED, PREDICTED, DIFF, RELERROR, ALL, NON, TREE;
	};

	// prediction tree algorithms
	public static enum Trees {
		SEQUOIA, SONG, MYTREE, DAE
	}

	// topologies
	public static enum Topologies {
		PERFECT, REALKNOWN, DSLCONNECTIONS, PLANETLAB, SUBSETPL
	}
	
	public static void main(String[] args) throws Exception {
		Locale.setDefault(Locale.US);

		// simulate cli arguments
		String s = "";
		s = "-to planetlab -tr sequoia"; 
		if (s.length() > 0 && args.length <= 0) {
			System.err.println("!!! Arguments overridden !!!");
			args = s.split(" ");
		}

		// arguments and parameters
		prepareCliArgs();
		CommandLineParser cliParser = new PosixParser();
		cli = cliParser.parse(options, args);
		if (args.length == 0 || cli.hasOption(HELP)) {
			HelpFormatter hf = new HelpFormatter();
			hf.printHelp("Master", options);
			System.exit(0);
		} else {
			processCliArgs();
		}
		
		// prepare filename components
		String fn_tree = cli_tree.toString().toLowerCase();
		String fn_topo = cli_topo.toString().toLowerCase();
		fn_topo = (cli_useclique && fn_topo.equals("planetlab")) ? fn_topo + "-clique" : fn_topo;
		fn_topo = cli_symratio != -1 ? fn_topo + "-symratio"+cli_symratio : fn_topo;
		String fn_pass = cli_iter > 1 ? "multi" : "single";
		String fn_prefix = String.format("%s-%s-%s-", fn_tree, fn_topo, fn_pass);

		// execution modes
		if (cli_treeness) {
			exec_treeness(fn_topo);
		} else if (cli_symmetry) {
			exec_symmetry(fn_topo);
		} else {
			if (cli_iter > 1) 
				exec_multi(fn_prefix);
			else 
				exec_single(fn_prefix);
		}
	}

	@SuppressWarnings("rawtypes")
	private static void exec_single(String filename_prefix) throws IOException {
		String testTitle = String.format("Testing %s @ %s", cli_topo, cli_tree);
		println(testTitle + notices());
		println("### initializing topology ... ");
		ITopology topology = Helper.getTopology(cli_topo);
		println("### embedding tree ... ");
		IPredictionTree tree = Helper.getTree(cli_tree, topology);
		println("### testing tree ... ");
		TreeTester test = new TreeTester(topology, tree);

		// cdf print / cdf to file
		if (cli_cdf) {
			CDF cdf = test.getRelErrorCDF();
			if(cli_toFile) {
				String filename = filename_prefix + "cdf";
				Helper.writeToFile(CDF_PATH, filename, cdf.toString());
				System.err.println("CDF written to \"" + CDF_PATH + filename + "\"");
			} else {
				System.out.println(cdf);
			}
		} else {
			TestResult result = test.getResults();
			System.out.println("Link error statistics: ");
			outputResults(result);
		}
	}

	private static void exec_multi(String filename_prefix) throws IOException {
		String testTitle = String.format("Testing %s @ %s", cli_topo, cli_tree);
		println(testTitle + notices());
		TestResult result = experiment(cli_tree, cli_topo, cli_iter);

		if (cli_cdf) {
			
			double[] values = result.getRelErrors().getValues();
			List<Double> list = new ArrayList<Double>();
			for(int i=0; i<values.length; i++)
				list.add(values[i]);
			
			CDF cdf = new CDF(list, CDF_ERROR_MAX);
			if(cli_toFile ) {
				String filename = filename_prefix + "relerror";
				Helper.writeToFile(CDF_PATH, filename, cdf.toString());
				System.err.println("CDF written to \"" + CDF_PATH + filename + "\"");
			} else {
				System.out.println(cdf);
			}
		} else {
			System.out.println("Tree error statistics: ");
			outputResults(result);
		}
	}

	private static void outputResults(TestResult result) {
		double meanRelError = result.getRelErrors().getMean();
		double errorPercentile = result.getRelErrors().getPercentile(PERCENTILE);
		double measRatio = result.getMeasurementRatio().getMean();
		
		if(cli_getavgerr) {
			double errorStdev = result.getRelErrors().getStandardDeviation();
			outputNodescaleSymratioStdev(meanRelError,errorStdev);
		} else if(cli_getavgmeas) {
			double measStdev = result.getMeasurementRatio().getStandardDeviation();
			outputNodescaleSymratioStdev(measRatio, measStdev);
		} else {
			System.out.printf("Mean rel. error: %s %s-percentile %s\nMeasurement ratio: %s\n",meanRelError,PERCENTILE,errorPercentile, measRatio);
		}
	}
	
	private static void outputNodescaleSymratioStdev(double value, double std) {
		if(cli.hasOption(NODES)) 
			System.out.print(cli_nodescale+";"+value);
		else if(cli.hasOption(SYMRATIO))
			System.out.print(cli_symratio+";"+value);
		else 
			System.out.print(value);
		
		if(cli_stdev)
			System.out.print(";"+std);
		
		System.out.println();
	}
	
	private static void exec_symmetry(String fn_topo) throws IOException {
		ITopology topology = Helper.getTopology(cli_topo);
		CDF cdf = topology.getSymmetry();
		System.out.println("Symmetry CDF of " + cli_topo + " (ratio; Phi)\n");
		String filename = "symmetry-" + fn_topo;
		Helper.writeToFile(CDF_PATH, filename, cdf.toString());
		System.err.println("CDF written to \"" + filename + "\"");
		System.err.println("  mean: " + cdf.getMean());		
	}

	private static void exec_treeness(String fn_topo) throws IOException {
		ITopology topology = Helper.getTopology(cli_topo);
		CDF cdf = topology.getTreeness();
		System.out.println("Treeness CDF of " + cli_topo + " (relError; Phi)\n");
		String filename = "treeness-" + fn_topo;
		Helper.writeToFile(CDF_PATH, filename, cdf.toString());
		System.err.println("CDF written to \"" + filename + "\"");
		System.err.println("  mean: " + cdf.getMean());
	}

	// Actual execution of Tests 
	@SuppressWarnings("rawtypes")
	private static TestResult experiment(Trees tree, Topologies topology, int iter) {
		TestResult result = new TestResult();
		
		for (int i = 0; i < iter; i++) {
			println("### initializing topology ... ");
			ITopology topo = Helper.getTopology(topology);
			println("### embedding ... ");
			IPredictionTree predTree = Helper.getTree(tree, topo);
			println("### testing ... ");
			TreeTester tester = new TreeTester(topo, predTree);
			
			// store valid(!) results
			if (tester.isValid()) {
				double error = tester.getResults().getRelErrors().getMean();
				if(Double.isInfinite(error) || Double.isNaN(error))
					continue;
				result.addError(error);
				double measurements = tester.getResults().getMeasurementRatio().getMean();
				if(Double.isInfinite(measurements) || Double.isNaN(measurements))
					continue;
				result.addMeasRatio(measurements);
				double embRatio = tester.getResults().getEmbeddingRatio().getMean();
				if(Double.isInfinite(embRatio) || Double.isNaN(embRatio))
					continue;
				result.addEmbRatio(embRatio);
			}
		}
		
		if (topology.equals(Topologies.PLANETLAB)) 
			println("");

		return result;
	}
	
	private static void println(String s) {
		if(cli_verbose)
			System.err.println(s);
	}

	// Displays important properties of current testrun 
	private static String notices() {
		String s = ": ";
		if (cli_iter > 1)
			s += cli_iter + "x, ";
		if (cli_errorlimit > 0)
			s += "error-limit=" + cli_errorlimit + ", ";
		if (cli_useclique)
			s += "clique-active, ";
		if (cli_dlabelmod)
			s += "distance label modification, ";
		if (cli_optimize)
			s += "anchor-selection with optimization, ";
		if (cli_nodescale > 0 && cli_topo.equals(Topologies.DSLCONNECTIONS.toString()))
			s += "node scale active: " + cli_nodescale + ", ";
		if (cli_tree.equals(Trees.SEQUOIA))
			s += "k="+cli_sequoiak + ", ";
		if(cli_symratio != -1)
			s += "zeta: " + cli_symratio + "; ";
		if(cli_random)
			s += "fixed random seed; ";
		return s;
	}

	// possible CLI arguments
	private static final String ITER = "iterations";
	private static final String TOPO = "topology";
	private static final String TREE = "tree";
	private static final String HELP = "help";
	private static final String OUTPUT = "output";
	private static final String CSV = "csv";
	private static final String CDF = "cdf";
	private static final String LIMIT = "limit";
	private static final String CLIQUE = "clique";
	private static final String FILES = "file";
	private static final String DLMOD = "dlmod";
	private static final String TREENESS = "treeness";
	private static final String SYMMETRY = "symmetry";
	private static final String NODES = "nodes";
	private static final String VERBOSE = "verbose";
	private static final String DEBUG = "debug";
	private static final String OPTIMIZE = "optimize";
	private static final String SYMRATIO = "symratio";
	private static final String AVGERROR = "avgerror";
	private static final String AVGMEAS = "avgmeas";
	private static final String SEQUOIAERROR = "seqerr";
	private static final String SEQUOIAK = "k";
	private static final String RANDOMSEED = "seed";
	private static final String STDDEV = "std";
	
	// (default) CLI argument values
	public static Trees cli_tree = Trees.SEQUOIA; // default tree
	static Topologies cli_topo = Topologies.PERFECT; // default topology
	static Output cli_output = Output.NON; // output mode
	static int cli_iter = 1; // amount of cycles
	static int cli_errorlimit = 0; // activate error limitation
	public int cli_seed = -1; // seed for random instance 
	static boolean cli_csv = false; // csv output
	static boolean cli_cdf = false; // calculate cdf
	static boolean cli_texOutputMode = false; // write cdf to file
	static boolean cli_useclique = false; // planetlab max clique
	static boolean cli_treeness = false; // treeness mode switch
	static boolean cli_symmetry = false; // symmetry mode switch
	static boolean cli_getavgerr = false; // return average error value 
	static boolean cli_getavgmeas = false; // return average measurements value
	static boolean cli_stdev = false; // standard deviation
	static boolean cli_toFile = false; // output to file (not stdout)
	static boolean cli_random = false; // whether or not random seed has been used
	public static int cli_sequoiak = 15; // Sequoia k-value (k-Trees)
	public static int cli_nodescale = 1; // node count (not any topology)
	public static double cli_symratio = -1; // ratio of servers to private hosts 0.0-1.0
	public static boolean cli_verbose = false; // verbose switch
	public static boolean cli_debug = false; // debug switch
	public static boolean cli_dlabelmod = false; // distance label modification
	public static boolean cli_optimize = false; // optimization
	public static boolean cli_sequoiaErrorCalc = false; // Sequoia way of calculating error
	
	private static void prepareCliArgs() {
		options = new Options();
		options.addOption("h", HELP, false, "Show help");
		options.addOption("tr", TREE, true, "Select tree implementation [sequoia, song, mytree])");
		options.addOption("to", TOPO, true, "Select topology implementation [perfect, dsl, realknown, planetlab]");
		options.addOption("i", ITER, true, "Define test iterations [1-n]");
		options.addOption("o", OUTPUT, true, "Select which value to print [measured, predicted, diff, rel, all]");
		options.addOption(CSV, false, "Activate csv output");
		options.addOption(CDF, false, "Activate CDF of relative error (instead of min/max... table)");
		options.addOption("l", LIMIT, true, "Define relative error cutoff-threshold");
		options.addOption("c", CLIQUE, false, "Activate usage of clique of hosts for the PlanetLab topology (instead of full host list)");
		options.addOption("f", FILES, false, "Write output to file");
		options.addOption("d", DLMOD, false, "Activate DL::distanceTo() modification introduced by me in Songs algorithm.");
		options.addOption(TREENESS, false, "Calculate treeness of topology (represented as cdf)");
		options.addOption(SYMMETRY, false, "Calculate symmetry of topology (represented as cdf)");
		options.addOption(NODES, true, "Scale amount of nodes of dataset (DSLConnections topology only)");
		options.addOption("v", VERBOSE, false, "Increase verbosity");
		options.addOption("dbg",DEBUG,false,"Activate debug output");
		options.addOption(OPTIMIZE, false, "Optimize DAE anchor selection algorithm");
		options.addOption(SYMRATIO,true,"Select asymmetry in the range of 0.0 to 1.0 (0.0 = symmetric Servers only, 1.0 = asymmetric private Hosts only");
		options.addOption(AVGERROR,false,"Return avg. error value only instead of min/max/avg/std table");
		options.addOption(AVGMEAS,false,"Return avg. measurements value only instead of min/max/avg/std table");
		options.addOption(STDDEV,false,"Return standard deviation of rel. error.");
		options.addOption(SEQUOIAERROR,false,"Calculate error the \"Sequoia way\", i.e. error with respect to average value");
		options.addOption(SEQUOIAK,true,"Set k-value for Sequoia tree embedding (default k=15).");
		options.addOption(RANDOMSEED,true,"Set seed used for the java.util.Random instance.");
	}

	private static void processCliArgs() {
		if (cli.hasOption(ITER)) {
			cli_iter = Integer.parseInt(cli.getOptionValue(ITER));
			cli_iter = cli_iter > 0 ? cli_iter : 1;
		}
	
		if (cli.hasOption(TREE)) {
			cli_tree = Trees.valueOf(cli.getOptionValue(TREE).toUpperCase());
			if(cli_tree.equals(Trees.DAE))
				cli_tree = Trees.MYTREE;
		}
		
		if (cli.hasOption(NODES)) {
			cli_nodescale = Integer.parseInt(cli.getOptionValue(NODES));
			cli_nodescale = cli_nodescale < 1 ? 1 : cli_nodescale;
		}
		
		if(cli.hasOption(SEQUOIAK )) {
			cli_sequoiak = Integer.parseInt(cli.getOptionValue(SEQUOIAK));
			cli_sequoiak = cli_sequoiak < 1 ? 1 : cli_sequoiak;
		}
	
		if (cli.hasOption(TOPO)) 
			cli_topo = Topologies.valueOf(cli.getOptionValue(TOPO).toUpperCase());
	
		if (cli.hasOption(CSV))
			cli_csv = true;
	
		if (cli.hasOption(CDF))
			cli_cdf = true;
	
		if (cli.hasOption(FILES))
			cli_texOutputMode = true;
	
		if (cli.hasOption(LIMIT))
			cli_errorlimit = Integer.parseInt(cli.getOptionValue(LIMIT));
	
		if (cli.hasOption(CLIQUE))
			cli_useclique = true;
	
		if (cli.hasOption(OUTPUT))
			cli_output = Output.valueOf(cli.getOptionValue(OUTPUT).toUpperCase());
	
		if (cli.hasOption(DLMOD))
			cli_dlabelmod = true;
	
		if (cli.hasOption(TREENESS))
			cli_treeness = true;
	
		if (cli.hasOption(SYMMETRY))
			cli_symmetry = true;
	
		if(cli.hasOption(SYMRATIO)) 
			cli_symratio = Double.parseDouble(cli.getOptionValue(SYMRATIO));
		
		if(cli.hasOption(AVGERROR)) 
			cli_getavgerr = true;
		
		if(cli.hasOption(AVGMEAS))
			cli_getavgmeas = true;
		
		if (cli.hasOption(VERBOSE))
			cli_verbose = true;
		
		if(cli.hasOption(DEBUG))
			cli_debug = true;
	
		if (cli.hasOption(OPTIMIZE))
			cli_optimize = true;
		
		if(cli.hasOption(SEQUOIAERROR))
			cli_sequoiaErrorCalc = true;
		
		if(cli.hasOption(STDDEV))
			cli_stdev = true;
		
		if(cli.hasOption(RANDOMSEED)) {
			cli_random = true;
			rand = new Random(Long.parseLong(cli.getOptionValue(RANDOMSEED)));
		} else {
			rand = new Random();
		}
	}

}
