/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia;

import java.util.ArrayList;
import java.util.List;

import sequoia.topology.ITopology;
import sequoia.tree.IPredictionTree;

/**
 * Takes a topology and a tree, and calculates the error statistics and the CDF
 * of prediction error.
 */
public class TreeTester {

	@SuppressWarnings("rawtypes")
	private IPredictionTree _tree;
	private ITopology _topology;

	private List<String> _hosts;
	private CDF _cdf = null;
	private TestResult _results = new TestResult();

	private boolean _test_valid = false;

	/**
	 * constructor taking a topology and a prediction tree and initiates the
	 * calculation of error statistics.
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public TreeTester(ITopology topology, IPredictionTree tree) {
		_topology = topology;
		_tree = tree;
		_hosts = _tree.getHostnames();
		double embedding_ratio = (double) _tree.getHostnames().size() / (double) _topology.getAllHosts().size();
		_results.addEmbRatio(embedding_ratio);
		calculate();
	}

	/**
	 * calculate abs/rel error
	 */
	public void calculate() {
		if(_hosts.size()<2)
			return; 
		
		// for every pair of two hosts, compare measured to predicted value
		for (String a : _hosts) {
			for (String b : _hosts) {
				if (a.equals(b))
					continue;
				
				// measured value is the requested link directly. However, when the Sequoia way 
				// of calculating error is activated, the link bandwidth is initially averaged.
				double measuredBandwidth = 0;
				try {
					// calculate error with respect to average measured value (Sequoia)
					if(Master.cli_sequoiaErrorCalc) {
						double measureda = _topology.getBandwidth(a, b, false);
						double measuredb = _topology.getBandwidth(b, a, false);
						if(measureda == 0 && measuredb != 0)
							measureda = measuredb;
						else if(measuredb == 0 && measureda != 0)
							measuredb = measureda;
						measuredBandwidth = (measureda+measuredb)/2;
					} else {
						measuredBandwidth = _topology.getBandwidth(a, b, false);
					}
					
					if(measuredBandwidth == 0)
						continue;
					
					double predictedBandwidth = _tree.predictBandwidth(a, b);
					double absError = Math.abs(measuredBandwidth - predictedBandwidth);
					double relError = absError / measuredBandwidth;
					
					// throw warnings, or store values
					if (Double.isInfinite(relError) || Double.isNaN(relError)) {
						System.err.printf("WARN: rel error %s>%s is NaN or Inf\n", a, b);
					} else {
						_results.addError(relError);
					}
				} catch(Exception e) {
					System.err.println(e.getMessage()+ " [...]");
				}
			}
		}
		
		_test_valid = true;
		_results.addMeasRatio(_tree.getMeasurementRatio());
	}
	
	public TestResult getResults() {
		return _results;
	}

	/**
	 * returns the pre-calculated CDF (Map<error, cumulative-share>)
	 */
	public CDF getRelErrorCDF() {
		if(_cdf == null) {
			double[] rerrors = _results.getRelErrors().getValues();
			// Guava needed:
			// List<Double> list = Doubles.asList(rerrors);
			List<Double> list = new ArrayList<Double>();
			for(int i=0; i<rerrors.length; i++)
				list.add(rerrors[i]);
			_cdf = new CDF(list,Master.CDF_ERROR_MAX);
		}
		return _cdf;
	}

	/**
	 * returns if test results are valid (no NaN and Inf)
	 */
	public boolean isValid() {
		return _test_valid;
	}

	/**
	 * returns general meta-information about testrun
	 */
	public String toString() {
		String s = "";
		s += _topology.getTopologyName().substring(3);
		s += " @ ";
		s += _tree.getClass().getSimpleName().substring(4);
		s += "\n";
		return s;
	}
}
