/**
 *
 * Copyright 2013 Stefan Brenner, Sven Schober
 *
 * This file is part of DAE.
 *
 * DAE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * DAE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with DAE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package sequoia.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sequoia.Master;
import sequoia.topology.ITopology;

/**
 * Implementation of Song's approach for prediction tree creation
 * 
 * @author stefan
 * 
 */
public class TreeSong extends PredictionTree<Host> {

	private static final double NA_BW = 0;
	private static final double NA_DIST = Master.CONST;
	protected List<Host> _hosts = new ArrayList<Host>();
	protected Host _base;
	private String _first;
	private Map<String, List<String>> _measured = new HashMap<String, List<String>>();

	public TreeSong(ITopology top) {
		_topology = top;
	}

	public void embed(String newHost) {
		if (_hosts.contains(newHost))
			return;

		if (_hosts.size() == 0) {
			firstJoin(newHost);
		} else if (_hosts.size() == 1) {
			secondJoin(newHost);
		} else {
			anyJoin(newHost);
		}
	}

	protected void firstJoin(String hostname) {
		Host joiner = new Host(hostname);
		joiner.parent = null;
		joiner.label.add(joiner, 0, 0);
		_first = hostname;
		_hosts.add(joiner);
	}

	protected void secondJoin(String hostname) {
		Host joiner = new Host(hostname);
		joiner.parent = _hosts.get(0);
		joiner.parent.children.add(joiner);
		double measuredDistance = measureDistance(joiner, joiner.parent);
		if((Master.CONST - measuredDistance) < 1)
			throw new RuntimeException("Unusable distance between first and second node.");
		joiner.label.add(joiner.parent, 0, measuredDistance);
		joiner.label.add(joiner, 0, 0);
		_hosts.add(joiner);
	}

	protected void anyJoin(String hostname) {
		Host joiner = new Host(hostname);
		Host base = null;
		Host anchor = null;
	
		while (anchor == null) {
			int randIdx = (int) Math.round(Master.rand.nextDouble() * (_hosts.size() - 1));
			int idx = _hosts.size() <= 2 ? 0 : randIdx;
			base = _hosts.get(idx);
			anchor = getAnchor(base, joiner);
		}
	
		anchor = optimize(joiner);
		if (_base != null)
			base = _base;
	
		assert (anchor != null);
		assert (!anchor.equals(base));
	
		joiner = position(joiner, anchor, base);
		joiner.parent = anchor;
		joiner.parent.children.add(joiner);
		_hosts.add(joiner);
		
		if (Master.cli_verbose && _hosts.size() % 10 == 0)
			System.err.print(_hosts.size()+" ");
	}

	protected Host getAnchor(Host origin, Host joiner) {
		// FIRST PHASE: static base node
		Host basis = origin;
		Host anchor = null;
		double minError = Double.POSITIVE_INFINITY;
	
		while (true) {
			List<Host> candList = getCandidates(basis);
			for (Host anchor_cand : candList) {
				if (anchor_cand == null || anchor_cand.equals(origin) || anchor_cand.hostname.equals(_first))
					continue;
	
				Host tempPosition = position(joiner, anchor_cand, origin);
				double candError = minimizeError(tempPosition);
				if (candError < minError) {
					minError = candError;
					anchor = anchor_cand;
				}
			}
	
			if (anchor == basis || anchor == null)
				break;
	
			basis = anchor;
		}
		return anchor;
	}

	protected List<Host> getCandidates(Host b) {
		List<Host> list = new ArrayList<Host>(b.children);
		Collections.copy(list, b.children);
		list.add(b.parent);
		list.add(b);
		return list;
	}

	protected Host optimize(Host joiner) {
		// SECOND PHASE: dynamic base node
		double min = Double.POSITIVE_INFINITY;
		Host anchor = null;
		_base = null;
		List<String> measuredHosts = getMeasuredList(joiner.hostname);
		for (String base : measuredHosts) {
			Host basenode = getHost(new Host(base));
			for (String anc : measuredHosts) {
				if (base.equals(anc) || anc.equals(_first))
					continue;
	
				Host anchor_cand = getHost(new Host(anc));
				Host tempPosition = position(joiner, anchor_cand, basenode);
				double error = minimizeError(tempPosition);
				if (error < min) {
					min = error;
					_base = basenode;
					anchor = anchor_cand;
				}
			}
		}
		return anchor;
	}

	protected Host position(Host joiner, Host anchor, Host basenode) {
		double distance_jb = measureDistance(joiner, basenode);
		double bandwidth_ab = predictBandwidth(anchor.hostname, basenode.hostname);
		double distance_ab = transform(bandwidth_ab);
		double distance_ja = measureDistance(joiner, anchor);
		
		if((Master.CONST - distance_ja) < 1 || (Master.CONST - distance_jb) < 1 )
			throw new RuntimeException("Unusable joiner >< {base,anchor} distance");
		
		if((Master.CONST - distance_ab) < 1)
			throw new RuntimeException("Unusable distance anchor >< base");
		
		double gromov = gromovProduct(distance_jb, distance_ab, distance_ja);
		double offset = distance_ab - gromov;
		double distance = distance_jb - gromov;
	
		Host position = new Host(joiner.hostname);
		position.label = anchor.label.clone();
		position.label.updateLast(offset, distance);
		position.label.add(joiner, 0, 0);
	
		return position;
	}

	protected double minimizeError(Host tempHost) {
		double sum = 0;
		int count = 0;
		List<String> measuredHosts = getMeasuredList(tempHost.hostname);
		for (String candidate : measuredHosts) {
			if (candidate.equals(tempHost.hostname))
				continue;
	
			double measuredDistance = measureDistance(tempHost.hostname, candidate);
			double measuredBandwidth = transform(measuredDistance);
			Host host = getHost(new Host(candidate)); 
			double predictedDistance = tempHost.label.distanceTo(host.label);
			double predictedBandwidth = transform(predictedDistance);
			double relError = Math.abs(measuredBandwidth - predictedBandwidth) / measuredBandwidth;
			sum += relError;
			count++;
		}
		double error = sum / count;
		error = Double.isNaN(error) ? Double.POSITIVE_INFINITY : error;
		return error;
	}

	@Override
	@SuppressWarnings("unchecked")
	public double predictBandwidth(String a, String b) {
		DistanceLabel<Host> la = getDLabel(a);
		DistanceLabel<Host> lb = getDLabel(b);
		double distance = la.distanceTo(lb);
		double bandwidth = transform(distance);
		return bandwidth;
	}

	protected double measureDistance(Host a, Host b) {
		return measureDistance(a.hostname, b.hostname);
	}

	protected double measureDistance(String a, String b) {
		double bandwidth = measureBandwidth(a,b);
		if(bandwidth == 0) return NA_DIST;
		double distance = transform(bandwidth);
		return distance;
	}

	protected double measureBandwidth(String a, String b) {
		double bandwidth_forth = (_topology.getBandwidth(a, b));
		double bandwidth_back = (_topology.getBandwidth(b, a));
		if(bandwidth_forth == 0 && bandwidth_back == 0)
			return NA_BW;
		countMeasurement(a, b);
		if(bandwidth_forth == 0 && bandwidth_back != 0)
			bandwidth_forth = bandwidth_back;
		if(bandwidth_back == 0 && bandwidth_forth != 0)
			bandwidth_back = bandwidth_forth;
		double bandwidth = (bandwidth_forth + bandwidth_back) / 2;
		return bandwidth;
	}

	protected void countMeasurement(String a, String b) {
		if(!_measured.containsKey(a))
			_measured.put(a, new ArrayList<String>());
		if(!_measured.get(a).contains(b))
			_measured.get(a).add(b);
		if(!_measured.containsKey(b))
			_measured.put(b, new ArrayList<String>());
		if(!_measured.get(b).contains(a))
			_measured.get(b).add(a);
	}
	
	public double getMeasurementRatio(){
		double measurements = 0;
		List<String> hostnames = getHostnames();
		// for all tree hosts
		for(Host host : _hosts) {
			// count measurements originating from them to other hosts available on the tree
			if(_measured.containsKey(host.hostname)) {
				List<String> measuredList = _measured.get(host.hostname);
				for(String destinationHostname : measuredList) {
					if(hostnames.contains(destinationHostname))
						measurements++;
				}
			}
		}
		double size = _hosts.size();
		double links = (size * (size - 1));
		double ratio = measurements / links;
		return ratio;
	}

	protected List<String> getMeasuredList(String hostname) {
		List<String> list = new ArrayList<String>(_measured.get(hostname));
		Collections.copy(list,_measured.get(hostname));
		return list;
	}

	public List<String> getHostnames() {
		ArrayList<Host> hosts = new ArrayList<Host>(_hosts);
		Collections.copy(hosts, _hosts);
		Collections.sort(hosts, new HostLabelComparator());
		List<String> list = new ArrayList<String>();
		for (Host h : hosts)
			list.add(h.hostname);
		return list;
	}

	@Override
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public DistanceLabel getDLabel(String host) {
		return getHost(new Host(host)).label;
	}

	protected Host getHost(Host host) {
		int idx = _hosts.indexOf(host);
		if (idx >= 0) {
			return _hosts.get(idx);
		} else {
			throw new RuntimeException("Host not found: getHost(" + host + ")");
		}
	}

	protected String hostnameFormat(String s) {
		return s.length() > 14 ? s.substring(0,14) : s;
	}

	@Override
	public List<Host> getHosts() {
		ArrayList<Host> hosts = new ArrayList<Host>(_hosts);
		Collections.copy(hosts, _hosts);
		return hosts;
	}

	public double transform(double v) {
			if(v > Master.CONST) 
				System.err.println("Transform value greater than CONST "+v);
			if(Math.abs(v) < 0.1) 
				System.err.println("Transform value very small "+v);
			return Master.CONST / v;
		}

	public String toString() {
		ArrayList<Host> hosts2 = new ArrayList<Host>(_hosts);
		Collections.copy(hosts2, _hosts);
		Collections.sort(hosts2, new HostLabelComparator());
		String result = "Tree Structure of "+ hosts2.size() + " nodes\n";
		for (Host h : hosts2) 
			result += String.format("%10s\t%s\n", hostnameFormat(h.toString()), h.label);
	
		ArrayList<String> hosts = (ArrayList<String>) getHostnames();
		Collections.sort(hosts);
		return result;
	}
}
